DECLARE FUNCTION INTST! (It!, Ix!, Iy)
DECLARE SUB radix1 (y#, n!, sv#, d#)
DECLARE SUB radix2 (rd#, n!, d#, k!)
REM
REM * * * * * * * * * * * * * * * * * * * * * * *
REM * Program to determine zygosity             *
REM * Jinghua Zhao, Pak Sham       10/16/1996   *
REM * * * * * * * * * * * * * * * * * * * * * * *
REM
REM INPUT: TWIN.LOC and TWIN.DAT
REM        TWIN.LOC -- the locus file
REM             nloci                      -- # of loci
REM             # alleles allele frequencies -- locus 1
REM             # alleles allele frequencies -- locus 2
REM               ...     ... ...               ...
REM             # alleles allele frequencies -- locus n
REM
REM        TWIN.DAT -- the genotype file
REM             Twin 1 ID  locus 1 locus 2 ...
REM             Twin 2 ID  locus 1 locus 2 ...
REM             ...         ... ...
REM             ...         ... ...
REM             ID are in number
REM
REM OUTPUT: expected and observed twin zygosity
REM
REM LIMITATION: max # alleles = 30
REM             change MAXALL to increase capability
REM             max # Chi-square statistics simulations = 1000
REM             change MAXSIM to increase # tables to get
REM             MS-DOS only
REM   radix1 -- max # loci = 33 (accurate to machine precision)
REM   radix2 -- No limit for # loci
REM
REM USAGE:
REM        QBASIC TWIN to see in the screen
REM        QBASIC TWIN > outfile to save in a file
REM
REM        TWIN > outfile for TWIN.EXE
REM
REM  CHECK: see CHECK.IBS for individual pair information
REM
DEFDBL A-H, O-Z
REM  ---- CHANGE VALUES IF NECESSARY -----------------------------
MAXALL = 30: MAXSIM = 1000
REM  -------------------------------------------------------------
PRINT " * * * * * * * * * * * * * * * * * * * * * * * *"
PRINT " *   Program to determine twin zygosity        *"
PRINT " *   Jinghua Zhao, Pak Sham       10/16/1996   *"
PRINT " * * * * * * * * * * * * * * * * * * * * * * * *"
CLS

ON ERROR GOTO errline

locfile$ = "twin.loc": datfile$ = "twin.dat": prolix$ = "check.ibs"
OPEN locfile$ FOR INPUT AS #1

INPUT #1, nloci
maxn = 3 ^ nloci: nloci2 = 2 * nloci: n = nloci

DIM n(nloci), p(nloci, MAXALL), q(nloci, 0 TO 2)
DIM pibs(0 TO nloci2), pt(0 TO nloci2), cpt(0 TO nloci2), nibs(0 TO nloci2)
DIM sok(0 TO nloci2), sek(0 TO nloci2), nobs(0 TO nloci2), expibs(0 TO nloci2)
DIM SHARED sv(0 TO n), f(nloci), d(0 TO n)
DIM locus1(nloci, 2), locus2(nloci, 2)

i = 1
ibssum = 0
WHILE NOT EOF(1)
 INPUT #1, n(i)
 sump = 0: PI2 = 0
 FOR j = 1 TO n(i)
     INPUT #1, p(i, j)
     sump = sump + p(i, j)
     PI2 = PI2 + p(i, j) ^ 2
 NEXT
 IF ABS(sump - 1) > .001 THEN ERROR 1
 P2 = 0: P1 = 0: P0 = 0
 FOR IA = 1 TO n(i)
 FOR IB = 1 TO n(i)
 FOR IC = 1 TO n(i)
 FOR ID = 1 TO n(i)
     PV = p(i, IA) * p(i, IB) * p(i, IC) * p(i, ID)
     s1 = INTST(IA, IC, ID) + INTST(IB, IC, ID)
     s2 = INTST(IC, IA, IB) + INTST(ID, IA, IB)
     s = s1 + s2
     IF (s = 4) THEN P2 = P2 + PV
     IF (s = 3 OR s = 2) THEN P1 = P1 + PV
     IF (s = 0) THEN P0 = P0 + PV
 NEXT ID, IC, IB, IA
 q(i, 0) = .25 * P0
 q(i, 1) = .25 * P1 + .5 * (1 - PI2)
 q(i, 2) = .25 * P2 + .5 * PI2 + .25
 i = i + 1
 ibssum = ibssum + (2 * .25 + (1 + PI2) * .5 + (2 * P2 + P1) * .25)
WEND

CLOSE #1

PRINT
PRINT USING "Number of loci = ###  "; nloci

FOR i = 0 TO nloci2
    pibs(i) = 0: nibs(i) = 0
NEXT
prod = 1
FOR j = 1 TO nloci
    f(j) = q(j, 0)
    prod = prod * f(j)
NEXT
pibs(0) = prod
REM array sv is for subroutine radix1, unnecessary for radix2
REM remove the following REMark statements for radix1
FOR i = 0 TO n
    sv(i) = 3 ^ i
NEXT
FOR i = 1 TO maxn - 1
REM x = i
REM CALL radix1(x, n, sv, d)
REM FOR j = n TO 1 STEP -1
REM     PRINT USING "## "; d(j);
REM NEXT
REM PRINT
    CALL radix2(2, n, d, 0)
REM remove remark to see conversion
REM FOR j = n - 1 TO 0 STEP -1
REM     PRINT USING "## "; d(j);
REM NEXT
REM PRINT
    prod = 1
    isum = 0
    FOR j = 1 TO nloci
REM     f(j) = q(j, d(j))
        f(j) = q(j, d(j - 1))
        prod = prod * f(j)
REM     isum = isum + d(j)
        isum = isum + d(j - 1)
    NEXT
    pibs(isum) = pibs(isum) + prod
NEXT

PRINT : PRINT "IBS distribution for DZ twins"
PRINT STRING$(47, "-")
PRINT "  IBS   probability   truncated probability"
PRINT STRING$(47, "-")

mean = 0: sd = 0
FOR i = 0 TO nloci2
    mean = mean + pibs(i) * i
    sd = sd + pibs(i) * i * i
    PRINT USING " ###   ###.####### "; i; pibs(i);
    IF (i < nloci2) THEN
       pt(i) = pibs(i) / (1 - pibs(nloci2))
       PRINT USING "   ###.#######"; pt(i)
    ELSE
       pt(i) = 0
       PRINT "   *** truncated **** "
    END IF
NEXT
sd = SQR(sd - mean ^ 2)

PRINT STRING$(47, "-")
PRINT USING " Mean = ####.##,   standard deviation = ####.## "; mean; sd
PRINT : PRINT
PRINT "Zygosity probabilities and IBS status for each pair"
PRINT STRING$(67, "-")
PRINT "  Pair     ID       P(MZ)        P(DZ)         IBS      assignment"
PRINT STRING$(67, "-")

OPEN datfile$ FOR INPUT AS #2
OPEN prolix$ FOR OUTPUT AS #3

PRINT #3, DATE$; "   "; TIME$
PRINT #3,

ip = 0
WHILE NOT EOF(2)
 ip = ip + 1
 INPUT #2, id1
 IF EOF(2) THEN ERROR 2
 FOR i = 1 TO nloci
     INPUT #2, locus1(i, 1), locus1(i, 2)
 NEXT
 INPUT #2, ID2
 IF EOF(2) THEN ERROR 2
 FOR i = 1 TO nloci
     INPUT #2, locus2(i, 1), locus2(i, 2)
 NEXT
 PRINT USING "#####  ######"; ip; id1;
 PRINT #3, USING "Pair ####"; ip
 PRINT #3, USING "ID = ####"; id1;
 PRINT #3, " LOCUS      ALLELES     IBS  COUNTS      OR"
 ORs = 1
 ibssum = 0
 FOR i = 1 TO nloci
     L11 = locus1(i, 1): L12 = locus1(i, 2)
     L21 = locus2(i, 1): L22 = locus2(i, 2)
     PI = p(i, L11)
     PJ = p(i, L12)
     s1 = INTST(L11, L21, L22) + INTST(L12, L21, L22)
     s2 = INTST(L21, L11, L12) + INTST(L22, L11, L12)
     s = s1 + s2
     hom = (s = 4) AND (L11 = L12)
     het = (s = 4) AND ((L11 <> L12) OR (L21 <> L22))
     IF (s = 4) THEN IBS = 2
     IF (s = 3 OR s = 2) THEN IBS = 1
     IF (s = 0) THEN IBS = 0
     IF hom THEN ORs = ORs / (.25 + .5 * PI + .25 * PI ^ 2)
     IF het THEN ORs = ORs / (.25 + .25 * (PI + PJ) + .5 * (PI * PJ))
     IF (s < 3.5) THEN ORs = ORs * 0
     PRINT #3, USING "          ####    "; i;
     PRINT #3, USING "## ##  ## ##    ##   "; L11; L12; L21; L22; IBS;
     PRINT #3, USING "##   ########.##"; s; ORs
     ibssum = ibssum + IBS
 NEXT
 OR1 = ORs / (1 + ORs): OR2 = 1 / (1 + ORs)
 PRINT USING "      #.######     #.######     #####"; OR1; OR2; ibssum;
 IF (OR1 > OR2) THEN PRINT "        MZ" ELSE PRINT "        DZ"
 nibs(ibssum) = nibs(ibssum) + 1
WEND
PRINT STRING$(67, "-")

PRINT : PRINT
PRINT "      Goodness of fit test (GFT)"
PRINT "between observed (O) and expected (E) IBS"
PRINT STRING$(44, "-")
PRINT " IBS    O       E       (O-E)^2/E"
PRINT STRING$(44, "-")
s = 0: c = 0
FOR i = 0 TO nloci2 - 1
    expibs(i) = pt(i) * (ip - nibs(nloci2))
    c = c + pt(i)
    cpt(i) = c
    chisq = 0
    IF (expibs(i) > 0) THEN chisq = (nibs(i) - expibs(i)) ^ 2 / expibs(i)
    PRINT USING " ###  ###  #####.### #####.#####"; i; nibs(i); expibs(i); chisq
    s = s + chisq
NEXT
PRINT STRING$(44, "-")
PRINT USING "Pearson Chi-Square = #######.##, DF = ###"; s; nloci2 - 1
GOTO nocollapse
REM
REM   ZHAO JH, 10/22/96 (UK)
REM
PRINT "      After combining adjacent groups"
PRINT "         for expected IBS < 5"
PRINT STRING$(44, "-")
PRINT " IBS    O       E       (O-E)^2/E"
PRINT STRING$(44, "-")
j = 0
FOR i = 0 TO nloci2 - 1
    so = 0: se = 0
    k = 0
    WHILE (se < 5 AND k + i <= nloci2 - 1)
         so = so + nibs(k + i)
         se = se + expibs(k + i)
         k = k + 1
    WEND
    i = i + (k - 1)
    j = j + 1
    sok(j) = so
    sek(j) = se
NEXT
IF (j = 1) THEN
    chisq = (sok(j) - sek(j)) ^ 2 / sek(j)
    t = chisq
    PRINT USING " ###  ###  #####.### #####.#####"; j; sok(j); sek(j); chisq
    j = 0
ELSE
    IF (sek(j) < 5) THEN
       sek(j - 1) = sek(j - 1) + sek(j)
       sok(j - 1) = sok(j - 1) + sok(j)
       j = (j - 1)
    END IF
    t = 0
    FOR i = 1 TO j
        chisq = (sok(i) - sek(i)) ^ 2 / sek(i)
        t = t + chisq
        PRINT USING " ###  ###  #####.### #####.#####"; j; sok(i); sek(i); chisq
    NEXT
    j = j - 1
END IF
PRINT STRING$(44, "-")
PRINT USING "Pearson Chi-Square = #######.##, DF = ###"; t; j
PRINT
nocollapse:
RANDOMIZE TIMER
k = 0
FOR i = 0 TO MAXSIM - 1
    FOR j = 0 TO nloci2 - 1: nobs(j) = 0: NEXT
    FOR j = 1 TO ip
        x = RND: l = 0
        FOR j1 = 0 TO nloci2 - 1
            IF (x >= cpt(j1)) THEN l = l + 1
        NEXT
        nobs(l) = nobs(l) + 1
    NEXT
    u = 0
    FOR j = 0 TO nloci2 - 1
        so = nobs(j): se = expibs(j)
        IF (se <> 0) THEN
           chisq = (so - se) ^ 2 / se
           u = u + chisq
        END IF
    NEXT
    IF (u >= s) THEN k = k + 1
NEXT
u = k / MAXSIM
PRINT USING "Empirical p value based on ##### simulations = #.#####"; MAXSIM; u

highn = nibs(nloci2 - 1): highp = pt(nloci2 - 1)
alldz = ip - nibs(nloci2)
pn = (1 - highp) ^ alldz
s = pn
FOR i = 1 TO highn - 1
    xi = i
    pn = pn * highp / (1 - highp) * (alldz - xi + 1) / xi
    s = s + pn
NEXT
PRINT USING "Probability of ### or more twin pairs "; highn;
PRINT USING "sharing #### alleles IBS = #.######"; nloci2 - 1; 1 - s
PRINT
PRINT USING "Estimated number of MZ twin pairs (IBS =###)= #####"; nloci2; nibs(nloci2)
PRINT USING "Estimated number of DZ twin pairs (IBS<=###)= #####"; nloci2 - 1; ip - nibs(nloci2)
PRINT
PRINT "Expected number of DZ twin pairs falsely classified as MZ =";
PRINT USING "####.#####"; pibs(nloci2) / (1 - pibs(nloci2)) * alldz

CLOSE #2, #3
END

errline:
PRINT
PRINT USING "Error ###: "; ERR
SELECT CASE ERR
  CASE 1
    PRINT USING "Sum of allele frequencies in locus ## = #.###"; i; sump
    RESUME NEXT
  CASE 2
    PRINT USING "There's error in ### pair"; ip
    RESUME NEXT
  CASE ELSE
    PRINT "Unexpected error, ending program."
    END
END SELECT

FUNCTION INTST (It, Ix, Iy)
IF (It = Ix OR It = Iy) THEN INTST = 1 ELSE INTST = 0
END FUNCTION

SUB radix1 (y, n, sv, d)
REM routine for converting individual decimal number to trinary.
REM ZHAO JH 10/16/96 (UK)
REM
x = y
FOR i = 0 TO n
    d(i) = 0
NEXT
kmax = 0
FOR i = 1 TO n
    FOR k = 0 TO n
        IF x = sv(k) GOTO L0
        IF x > sv(k) AND x < sv(k + 1) GOTO L1
    NEXT
L0: IF kmax < k THEN kmax = k
    d(k + 1) = 1
    GOTO done
L1: IF kmax < k THEN kmax = k
REM change here to allow for other radix
    IF x >= 2 * sv(k) THEN j = 2 ELSE IF k = 0 THEN j = x ELSE j = 1
    x = x - j * sv(k)
    d(k + 1) = j
    IF x = 0 THEN GOTO done
NEXT
done:
REM PRINT "##="; y
sum = 0
FOR i = kmax TO 0 STEP -1
    sum = sum + d(i + 1) * sv(i)
REM PRINT USING "### #### ##### #####"; d(i + 1); sv(i); d(i + 1) * sv(i);sum
NEXT
END SUB

SUB radix2 (rd, n, d, k)
REM routine to express decimal sequence to (rd+1)-base numbers
REM Zhao JH, 10/21/96 (UK)
    IF d(k) < rd THEN
       d(k) = d(k) + 1
       GOTO ok
    ELSE
       d(k) = 0
       d(k + 1) = d(k + 1) + 1
       IF d(k + 1) <= rd GOTO ok
    END IF
    CALL radix2(rd, n, d, k + 1)
ok:
END SUB

